EXPORTS
    vaInitialize
    vaErrorStr
    vaTerminate
    vaQueryVendorString
    vaMaxNumProfiles
    vaMaxNumEntrypoints
    vaQueryConfigProfiles
    vaQueryConfigEntrypoints
    vaGetConfigAttributes
    vaProfileStr
    vaEntrypointStr
    vaConfigAttribTypeStr
    vaAcquireBufferHandle
    vaAssociateSubpicture
    vaBeginPicture
    vaCreateBuffer
    vaCreateConfig
    vaCreateContext
    vaCreateImage
    vaCreateSubpicture
    vaCreateSurfaces
    vaDeassociateSubpicture
    vaDeriveImage
    vaDestroyBuffer
    vaDestroyConfig
    vaDestroyContext
    vaDestroyImage
    vaDestroySubpicture
    vaDestroySurfaces
    vaEndPicture
    vaGetDisplayAttributes
    vaGetImage
    vaMapBuffer
    vaMapBuffer2
    vaMaxNumDisplayAttributes
    vaMaxNumImageFormats
    vaMaxNumSubpictureFormats
    vaPutImage
    vaQueryDisplayAttributes
    vaQueryImageFormats
    vaQuerySubpictureFormats
    vaQuerySurfaceAttributes
    vaQuerySurfaceStatus
    vaQueryVideoProcFilterCaps
    vaQueryVideoProcFilters
    vaQueryVideoProcPipelineCaps
    vaReleaseBufferHandle
    vaRenderPicture
    vaSetDisplayAttributes
    vaSetErrorCallback
    vaSetInfoCallback
    vaSetSubpictureGlobalAlpha
    vaSyncSurface
    vaUnmapBuffer
    vaSetDriverName
    vaSyncBuffer
    vaQueryConfigAttributes
    vaBufferSetNumElements
    vaQuerySurfaceError
    vaSetImagePalette
    vaSetSubpictureChromakey
    vaBufferInfo
    vaLockSurface
    vaUnlockSurface
    vaCreateMFContext
    vaMFAddContext
    vaMFReleaseContext
    vaMFSubmit
    vaCreateBuffer2
    vaQueryProcessingRate
    vaExportSurfaceHandle
    va_newDisplayContext
    va_newDriverContext
    vaMaxNumConfigAttributes
    vaBufferTypeStr
    vaAttachProtectedSession
    vaCopy
    vaCreateProtectedSession
    vaDestroyProtectedSession
    vaDetachProtectedSession
    vaProtectedSessionExecute
    vaSetSubpictureImage
    vaStatusStr
    vaSyncSurface2
    vaDisplayIsValid
    vaGetLibFunc
